@@ -13,8 +13,8 @@ implementation (like GTK, Qt, or CLI).
 # option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
 # the full text of the license.
 
-import glob, sys, os.path, optparse, traceback, locale, gettext, re
-import errno, zlib
+import glob, sys, os.path, optparse, traceback, locale, gettext, re, io
+import errno, zlib, gzip
 import subprocess, threading, webbrowser
 import signal
 import time
@@ -33,7 +33,7 @@ else:
     from configparser import ConfigParser
     PY3 = True
 
-__version__ = '2.20.10'
+__version__ = '2.20.11'
 
 
 def excstr(exception):
@@ -112,7 +112,7 @@ def thread_collect_info(report, reportfile, package, ui, symptom_script=None,
             report['Symptom'] = os.path.splitext(os.path.basename(symptom_script))[0]
         except StopIteration:
             sys.exit(0)
-        except Exception as e:
+        except Exception:
             apport.error('symptom script %s crashed:', symptom_script)
             traceback.print_exc()
             sys.exit(0)
@@ -243,7 +243,11 @@ class UserInterface:
             logind_session = None
         else:
             reports = apport.fileutils.get_new_reports()
-            logind_session = apport.Report.get_logind_session(os.getpid())
+            if PY3:
+                proc_pid_fd = os.open('/proc/%s' % os.getpid(), os.O_RDONLY | os.O_PATH | os.O_DIRECTORY)
+                logind_session = apport.Report.get_logind_session(proc_pid_fd=proc_pid_fd)
+            else:
+                logind_session = apport.Report.get_logind_session(os.getpid())
 
         for f in reports:
             if not self.load_report(f):
@@ -394,7 +398,17 @@ class UserInterface:
         mark the report for uploading.
         '''
         self.report = apport.Report('Hang')
-        self.report.add_proc_info(pid)
+        try:
+            self.report.add_proc_info(pid)
+        except ValueError as e:
+            if str(e) == 'invalid process':
+                self.ui_error_message(_('Invalid PID'),
+                                      _('The specified process ID does not exist.'))
+                sys.exit(1)
+            elif str(e) == 'not accessible':
+                self.ui_error_message(_('Not your PID'),
+                                      _('The specified process ID does not belong to you.'))
+                sys.exit(1)
         self.report.add_package_info()
         path = self.report.get('ExecutablePath', '')
         self.cur_package = apport.fileutils.find_file_package(path)
@@ -453,14 +467,20 @@ class UserInterface:
         # if PID is given, add info
         if self.options.pid:
             try:
-                with open('/proc/%s/stat' % self.options.pid) as f:
+                proc_pid_fd = None
+                if PY3:
+                    proc_pid_fd = os.open('/proc/%s' % self.options.pid, os.O_RDONLY | os.O_PATH | os.O_DIRECTORY)
+                    stat_file = os.open('stat', os.O_RDONLY | os.O_CLOEXEC, dir_fd=proc_pid_fd)
+                else:
+                    stat_file = '/proc/%s/stat' % self.options.pid
+                with io.open(stat_file) as f:
                     stat = f.read().split()
                 flags = int(stat[8])
                 if flags & PF_KTHREAD:
                     # this PID is a kernel thread
                     self.options.package = 'linux'
                 else:
-                    self.report.add_proc_info(self.options.pid)
+                    self.report.add_proc_info(pid=self.options.pid, proc_pid_fd=proc_pid_fd)
             except (ValueError, IOError, OSError) as e:
                 if hasattr(e, 'errno'):
                     # silently ignore nonexisting PIDs; the user must not close the
@@ -515,13 +535,18 @@ class UserInterface:
 
         if self.options.save:
             try:
-                with open(os.path.expanduser(self.options.save), 'wb') as f:
-                    self.report.write(f)
+                savefile = os.path.expanduser(self.options.save)
+                if savefile.endswith(".gz"):
+                    with gzip.open(savefile, 'wb') as f:
+                        self.report.write(f)
+                else:
+                    with open(os.path.expanduser(self.options.save), 'wb') as f:
+                        self.report.write(f)
             except (IOError, OSError) as e:
                 self.ui_error_message(_('Cannot create report'), excstr(e))
         else:
             # show what's being sent
-            allowed_to_report = apport.fileutils.allowed_to_report()
+            allowed_to_report = True
             response = self.ui_present_report_details(allowed_to_report)
             if response['report']:
                 self.file_report()
@@ -600,7 +625,7 @@ class UserInterface:
             return False
 
         # show what's being sent
-        allowed_to_report = apport.fileutils.allowed_to_report()
+        allowed_to_report = True
         response = self.ui_present_report_details(allowed_to_report)
         if response['report']:
             self.crashdb.update(self.options.update_report, self.report,
@@ -696,23 +721,23 @@ class UserInterface:
                 self.ui_error_message(_('Invalid problem report'), excstr(e))
             return True
         elif self.options.window:
-                self.ui_info_message('', _('After closing this message '
-                                           'please click on an application window to report a problem about it.'))
-                xprop = subprocess.Popen(['xprop', '_NET_WM_PID'],
-                                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-                (out, err) = xprop.communicate()
-                if xprop.returncode == 0:
-                    try:
-                        self.options.pid = int(out.split()[-1])
-                    except ValueError:
-                        self.ui_error_message(_('Cannot create report'),
-                                              _('xprop failed to determine process ID of the window'))
-                        return True
-                    return self.run_report_bug()
-                else:
+            self.ui_info_message('', _('After closing this message '
+                                       'please click on an application window to report a problem about it.'))
+            xprop = subprocess.Popen(['xprop', '_NET_WM_PID'],
+                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+            (out, err) = xprop.communicate()
+            if xprop.returncode == 0:
+                try:
+                    self.options.pid = int(out.split()[-1])
+                except ValueError:
                     self.ui_error_message(_('Cannot create report'),
-                                          _('xprop failed to determine process ID of the window') + '\n\n' + err)
+                                          _('xprop failed to determine process ID of the window'))
                     return True
+                return self.run_report_bug()
+            else:
+                self.ui_error_message(_('Cannot create report'),
+                                      _('xprop failed to determine process ID of the window') + '\n\n' + err)
+                return True
         else:
             return self.run_crashes()
 
@@ -1086,7 +1111,7 @@ class UserInterface:
                                                     args=(self.report, self.report_file, self.cur_package,
                                                           hookui, symptom_script, ignore_uninstalled))
                 icthread.start()
-                while icthread.isAlive():
+                while icthread.is_alive():
                     self.ui_pulse_info_collection_progress()
                     if hookui:
                         try:
@@ -1151,7 +1176,7 @@ class UserInterface:
                 bpthread = apport.REThread.REThread(target=self.report.search_bug_patterns,
                                                     args=(self.crashdb.get_bugpattern_baseurl(),))
                 bpthread.start()
-                while bpthread.isAlive():
+                while bpthread.is_alive():
                     self.ui_pulse_info_collection_progress()
                     try:
                         bpthread.join(0.1)
@@ -1172,7 +1197,7 @@ class UserInterface:
                 known_thread = apport.REThread.REThread(target=self.crashdb.known,
                                                         args=(self.report,))
                 known_thread.start()
-                while known_thread.isAlive():
+                while known_thread.is_alive():
                     self.ui_pulse_info_collection_progress()
                     try:
                         known_thread.join(0.1)
@@ -1190,7 +1215,7 @@ class UserInterface:
             # might damage the stack trace
             anonymize_thread = apport.REThread.REThread(target=self.report.anonymize)
             anonymize_thread.start()
-            while anonymize_thread.isAlive():
+            while anonymize_thread.is_alive():
                 self.ui_pulse_info_collection_progress()
                 try:
                     anonymize_thread.join(0.1)
@@ -1269,7 +1294,7 @@ class UserInterface:
         try:
             try:
                 subprocess.call(sudo_prefix + ['xdg-open', url])
-            except OSError as e:
+            except OSError:
                 # fall back to webbrowser
                 webbrowser.open(url, new=True, autoraise=True)
                 sys.exit(0)
@@ -1317,7 +1342,7 @@ class UserInterface:
         upthread = apport.REThread.REThread(target=self.crashdb.upload,
                                             args=(self.report, progress_callback))
         upthread.start()
-        while upthread.isAlive():
+        while upthread.is_alive():
             self.ui_set_upload_progress(__upload_progress)
             try:
                 upthread.join(0.1)
